#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <krb5.h>
#include <com_err.h>
#include "kadm5/admin.h"
#include "k5helper.h"
#ifdef KRB5PLUGIN
  #include "slapi-plugin.h"
#endif

//
int k5_logger(char *operation,char * msg) {
#ifdef KRB5PLUGIN
  slapi_log_error( SLAPI_LOG_FATAL, "krb5-plugin","%s: %s\n",operation,msg);
#else
  printf("%s: %s\n",operation,msg);
#endif
};

//
int k5_addprinc(char *principal, char *password) {
        char keytab[256],tgsname[256]; 
  	krb5_context context,context2;
        void *handle = NULL;
        long retval;
        kadm5_principal_ent_rec princ;
        memset(&princ, 0, sizeof(princ));
	kadm5_config_params params;
        memset(&params, 0, sizeof(params));
        kadm5_init_krb5_context(&context);
        kadm5_init_krb5_context(&context2);
        params.mask |= KADM5_CONFIG_REALM;
        if((params.realm = strstr (principal, "@")) && params.realm ) {
          params.realm++;
        } else {
          if (retval=krb5_get_default_realm(context, &params.realm)) {
            k5_logger("krb5_get_default_realm",(char *)error_message(retval));
            return retval;
          }
        }
        strcpy(keytab,KEYTAB_PREFIX);
        strncat(keytab,params.realm,200);
        strcat(keytab,".keytab");
        strcpy(tgsname,TGSKEY);
        strncat(tgsname,params.realm,200);
        if (retval=kadm5_init_with_skey(tgsname, keytab, KADM5_ADMIN_SERVICE, &params, KADM5_STRUCT_VERSION, KADM5_API_VERSION_2, NULL, &handle)) {
          k5_logger("addprinc: Keytable ",keytab);
          k5_logger("addprinc: TGS      ",tgsname);
	  k5_logger("kadm5_init_with_skey",(char *)error_message(retval));	  
	  return retval;
        }
        krb5_parse_name(context2, principal, &princ.principal);
        if (retval=kadm5_create_principal(handle, &princ, 1, password)) {
	  k5_logger("kadm5_create_principal",(char *)error_message(retval));	  
	  return retval;
        }
        kadm5_destroy(handle);
	krb5_free_principal(context, princ.principal);
        return 0;
}
//
int k5_cpw(char *principal, char *password) {
        char keytab[256],tgsname[256]; 
  	krb5_context context,context2;
        void *handle = NULL;
        long retval;
        kadm5_principal_ent_rec princ;
        memset(&princ, 0, sizeof(princ));
	kadm5_config_params params;
        memset(&params, 0, sizeof(params));
        kadm5_init_krb5_context(&context);
        kadm5_init_krb5_context(&context2);
        params.mask |= KADM5_CONFIG_REALM;
        if((params.realm = strstr (principal, "@")) && params.realm ) {
          params.realm++;
        } else {
          if (retval=krb5_get_default_realm(context, &params.realm)) {
            k5_logger("krb5_get_default_realm",(char *)error_message(retval));
            return retval;
          }
        }
        strcpy(keytab,KEYTAB_PREFIX);
        strncat(keytab,params.realm,200);
        strcat(keytab,".keytab");
        strcpy(tgsname,TGSKEY);
        strncat(tgsname,params.realm,200);
        if (retval=kadm5_init_with_skey(tgsname, keytab, KADM5_ADMIN_SERVICE, &params, KADM5_STRUCT_VERSION, KADM5_API_VERSION_2, NULL, &handle)) {
          k5_logger("cpw: Keytable ",keytab);
          k5_logger("cpw: TGS      ",tgsname);
	  k5_logger("kadm5_init_with_skey",(char *)error_message(retval));	  
	  return retval;
        }
        krb5_parse_name(context2, principal, &princ.principal);
        if (retval=kadm5_chpass_principal(handle, princ.principal, password)) {
	  k5_logger("kadm5_chpass_principal",(char *)error_message(retval));	  
	  return retval;
        }
        kadm5_destroy(handle);
	krb5_free_principal(context, princ.principal);
        return 0;
}
//
long krb5_nd_verify_creds(char *pname, char *passwd)
{
   slapi_log_error( SLAPI_LOG_FATAL, "krb5-plugin","Trying to authenticate principal %s\n",pname);
   krb5_context context;
   krb5_principal princ;
   krb5_get_init_creds_opt opts;
   krb5_creds creds;
   long retval;
   char *charToUP = 0L;
   if (!(pname && passwd)) return (1);
   if((charToUP = strstr (pname, "@")) && charToUP ) {
     charToUP++;
     for(; *charToUP; charToUP++) {
         *charToUP = toupper(*charToUP);
     }
   }
   if (retval=krb5_init_context(&context)) {
     k5_logger("krb5_init_context",(char *)error_message(retval));	  
     return retval;
   }

   if (retval=krb5_parse_name(context, pname, &princ)) {
     k5_logger("krb5_parse_name",(char *)error_message(retval));	  
     return retval;
   }
   krb5_get_init_creds_opt_init(&opts);
   krb5_get_init_creds_opt_set_tkt_life(&opts, 5*60);
   krb5_get_init_creds_opt_set_renew_life(&opts, 0);
   krb5_get_init_creds_opt_set_forwardable(&opts, 0);
   krb5_get_init_creds_opt_set_proxiable(&opts, 0);
   if (retval=krb5_get_init_creds_password(context, &creds, princ, passwd, NULL, NULL, 0, NULL, &opts)) {
     k5_logger("krb5_get_init_creds_password",(char *)error_message(retval));	  
     //slapi_log_error( SLAPI_LOG_FATAL, "krb5-plugin","%lu\n",retval);
   }
   return retval;
}
