/** BEGIN COPYRIGHT BLOCK
 * This Program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 * 
 * This Program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this Program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * In addition, as a special exception, Red Hat, Inc. gives You the additional
 * right to link the code of this Program with code not covered under the GNU
 * General Public License ("Non-GPL Code") and to distribute linked combinations
 * including the two, subject to the limitations in this paragraph. Non-GPL Code
 * permitted under this exception must only link to the code of this Program
 * through those well defined interfaces identified in the file named EXCEPTION
 * found in the source code files (the "Approved Interfaces"). The files of
 * Non-GPL Code may instantiate templates or use macros or inline functions from
 * the Approved Interfaces without causing the resulting work to be covered by
 * the GNU General Public License. Only Red Hat, Inc. may make changes or
 * additions to the list of Approved Interfaces. You must obey the GNU General
 * Public License in all respects for all of the Program code and other code used
 * in conjunction with the Program except the Non-GPL Code covered by this
 * exception. If you modify this file, you may extend this exception to your
 * version of the file, but you are not obligated to do so. If you do not wish to
 * provide this exception without modification, you must delete this exception
 * statement from your version and license this file solely under the GPL without
 * exception. 
 * 
 * 
 * Copyright (C) 2001 Sun Microsystems, Inc. Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

/************************************************************


 To test this plug-in function, stop the server, edit the dse.ldif file
 (in the config directory)
 and add the following lines before restarting the server :

dn: cn=KRB5 Bind,cn=plugins,cn=config
objectClass: top
objectClass: nsSlapdPlugin
objectClass: extensibleObject
cn: KRB5 Bind
nsslapd-pluginPath: /usr/lib64/dirsrv/plugins/libkrb5-plugin.so
nsslapd-pluginInitfunc: krb5plugin_init
nsslapd-pluginType: preoperation
nsslapd-pluginEnabled: on
nsslapd-plugin-depends-on-type: database
nsslapd-pluginId: krb5plugin
nsslapd-pluginVersion: 1.0.1
nsslapd-pluginVendor: Fedora Project
nsslapd-pluginDescription: Kerberos 5  pre-operation plugin (INFN)


 ************************************************************/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <krb5.h>
#include "slapi-plugin.h"
#include <stdarg.h>
#include "k5helper.h"
#define SLAPI_USERPWD_ATTR "userpassword"
#define MAXPRINC 16
int lowerstring(char * s);
int upperstring(char * s);

Slapi_PluginDesc bindpdesc = { "krb5-bind", "Fedora Project", "1.0.0",
	"Kerberos 5 bind pre-operation plugin (INFN)" };

static Slapi_ComponentId *plugin_id = NULL;


/* Pre-operation plug-in function */
int
krb5_bind( Slapi_PBlock *pb )
{
	char			*dn, *attrs[2] = { SLAPI_USERPWD_ATTR, NULL };
	int			method, rc = LDAP_SUCCESS;
	struct berval		*credentials;
	struct berval		**pwvals;
    	Slapi_DN		*sdn = NULL;
	Slapi_Entry		*e = NULL;
	Slapi_Attr		*attr = NULL;
        long              get_init_creds_error;

	/* Gets parameters available when processing an LDAP bind
	   operation. */
	if ( slapi_pblock_get( pb, SLAPI_BIND_TARGET, &dn ) != 0 ||
	    slapi_pblock_get( pb, SLAPI_BIND_METHOD, &method ) != 0 ||
	    slapi_pblock_get( pb, SLAPI_BIND_CREDENTIALS, &credentials ) != 0 ) {
		slapi_send_ldap_result( pb, LDAP_OPERATIONS_ERROR, 
			NULL, NULL, 0, NULL );
		return( 1 );
	}

	/* Check the authentication method */
	switch( method ) {
	case LDAP_AUTH_SIMPLE:
		/* First, get the entry specified by the DN. */
		sdn = slapi_sdn_new_dn_byref( dn );
		rc = slapi_search_internal_get_entry( sdn, attrs, &e,
			    plugin_id );
		slapi_sdn_free( &sdn );

		if ( rc != LDAP_SUCCESS ) {
			slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind",
					"Could not find entry %s (error %d)\n", 
					dn, rc );
                        // Autenticazione Anonima anche se la entry non esiste
                        // rc = LDAP_SUCCESS ;
                        // dn[0]=0;
			break;
		}

		if ( e != NULL ) {
			Slapi_Value	*credval, **pwvals;
			int		i, hint, valcount,nprinc;
                        struct berval   **auth_idval;
                        
		        char principal[MAXPRINC][255], password[255], realm[255];
			realm[0]=0;
			get_init_creds_error=1;
                        nprinc=0;
                        if ( slapi_entry_attr_find( e, "infnKerberosPrincipal", &attr ) == 0 ) {
                           Slapi_Value  * princ_val;
                           for ( hint = slapi_attr_first_value( attr, &princ_val ); hint != -1 && nprinc<MAXPRINC; hint = slapi_attr_next_value( attr, hint, &princ_val )) {
                              char * s;
			      s = (char *)slapi_value_get_string(princ_val);
                              strncpy(principal[nprinc],s,254); 
			      nprinc++;
                           }
                        }

                        rc=LDAP_INVALID_CREDENTIALS; 
                        if (nprinc) {
                           if (credentials->bv_val!=NULL) {
                             strncpy(password, credentials->bv_val,254);
                             for (i=0; i<nprinc; i++) {
                               get_init_creds_error = krb5_nd_verify_creds(principal[i], password);
                               if (!get_init_creds_error) {
                                  slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind","%s successfully authenticated\n",principal[i]);
                                  rc = LDAP_SUCCESS;
 				  break;
                               } else {
                                  slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind","Kerberos 5 authentication failed for the principal %s\n",principal[i]);
                                  rc = LDAP_INVALID_CREDENTIALS;
                               }
			     }
                           } else {
                             slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind","Kerberos 5 authentication failed. No password has been provided.\n");
                             //rc = LDAP_INVALID_CREDENTIALS;
                             rc = LDAP_SUCCESS;
                             dn[0]=0;
                           } 

                        }
                        		        
                        // 2529638918UL = Client not found in Kerberos database
                        // if  (rc != LDAP_SUCCESS  && (!(int)(get_init_creds_error-2529638918UL) || !strlen(principal[0]))) {

                        if  (rc != LDAP_SUCCESS) {
			  slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind", "Trying to authenticate %s against hash password\n",dn);
       			  if ( slapi_entry_attr_find( e, SLAPI_USERPWD_ATTR,
			      &attr ) != 0 || slapi_attr_get_numvalues( attr,
			      &valcount ) != 0 ) {
			    	  slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind",
				  	"Entry has no %s attribute values\n",
					SLAPI_USERPWD_ATTR );
				rc = LDAP_INVALID_CREDENTIALS;
				break;
			  }

			  credval = slapi_value_new_berval( credentials );
			  pwvals = (Slapi_Value **)slapi_ch_calloc( valcount,
			    sizeof( Slapi_Value * ));
			  i = 0;
			  for ( hint = slapi_attr_first_value( attr, &pwvals[i] );
			    hint != -1; hint = slapi_attr_next_value( attr,
			    hint, &pwvals[i] )) {
				++i;
			  }
			  if (strlen((char *)credval)) {
			    if ( slapi_pw_find_sv( pwvals, credval ) != 0 ) {
				slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind","Credentials are not correct for %s\n",dn);
				rc = LDAP_INVALID_CREDENTIALS;
			    } else {
                                slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind","Hash password Authentication OK for %s\n", dn );
                                rc =  LDAP_SUCCESS;
				if (strlen(principal[0]) && !strstr(dn,"ou=people,dc=infn,dc=it")) {
				  int krb5_perror;
                               	  slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind", "Trying to add the principal %s\n", principal);
			          krb5_perror = k5_addprinc(principal[0], password);
				  if (krb5_perror){
			  	    slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind", "ERROR: Principal %s has not been added.\n", principal);
				  } else {
			  	    slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind", "SUCCESS: Principal %s has been added.\n", principal);
				  }
				}
                            }
                          } else {
                            dn[0]=0;
                            rc = LDAP_SUCCESS;
                          }
			  slapi_value_free( &credval );
			  slapi_ch_free( (void **)&pwvals );
                        }
			if ( LDAP_SUCCESS != rc ) {
			  break;
			}
		} else {
			/* This should not happen. The previous section of code 
			   already checks for this case. */
			slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind",
				"Could find entry for %s\n", dn );
			rc = LDAP_NO_SUCH_OBJECT;
			break;
		}

		/* Set the DN and authentication method for the connection. */
		if ( slapi_pblock_set( pb, SLAPI_CONN_DN, 
			slapi_ch_strdup( dn ) ) != 0 ||
		     slapi_pblock_set( pb, SLAPI_CONN_AUTHMETHOD, 
			SLAPD_AUTH_SIMPLE ) != 0 ) {

			slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind",
				"Failed to set DN and method for connection\n" );
			rc = LDAP_OPERATIONS_ERROR;
			break;
		}

		/* Send a "success" result code back to the client. */
		slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind", 
			"Authenticated: %s\n", dn );
		rc = LDAP_SUCCESS;
		break;

	/* If NONE is specified, the client is requesting to bind anonymously.
	   Normally, this case should be handled by the server's front-end
	   before it calls this plug-in function.  Just in case this does
	   get through to the plug-in function, you can handle this by
	   sending a successful result code back to the client and returning
	   1.
	 */
	case LDAP_AUTH_NONE:
		slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind", 
			"Authenticating anonymously\n" );
		rc = LDAP_SUCCESS;
		break;

	/* This plug-in does not support any other method of authentication */
	case LDAP_AUTH_SASL:
             slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind","SASL Authentication OK\n");
             return (0);
	default:
		slapi_log_error( SLAPI_LOG_FATAL, "krb5_bind",
			"Unsupported authentication method requested: %d\n",
			method );
		rc = LDAP_AUTH_METHOD_NOT_SUPPORTED;
		break;
	}
	slapi_send_ldap_result( pb, rc, NULL, NULL, 0, NULL );
	return( 1 );
}


krb5_modify( Slapi_PBlock *pb )
{
	char			*dn, *attrs[2] = { SLAPI_USERPWD_ATTR, NULL };
	int			method, rc = LDAP_SUCCESS;
	struct berval		*credentials;
	struct berval		**pwvals;
    	Slapi_DN		*sdn = NULL;
	Slapi_Entry		*e = NULL;
	Slapi_Attr		*attr = NULL;


        //slapi_log_error( SLAPI_LOG_FATAL, "krb5_modify","MODIFY\n" );

	/* Gets parameters available when processing an LDAP bind
	   operation. */
	if ( slapi_pblock_get( pb, SLAPI_BIND_TARGET, &dn ) != 0 ||
	    slapi_pblock_get( pb, SLAPI_BIND_METHOD, &method ) != 0 ||
	    slapi_pblock_get( pb, SLAPI_BIND_CREDENTIALS, &credentials ) != 0 ) {
		slapi_send_ldap_result( pb, LDAP_OPERATIONS_ERROR,
			NULL, NULL, 0, NULL );
		return( 1 );
	}

	sdn = slapi_sdn_new_dn_byref( dn );
	rc = slapi_search_internal_get_entry( sdn, attrs, &e,plugin_id );
	slapi_sdn_free( &sdn );
	if ( rc != LDAP_SUCCESS ) {
	  slapi_log_error( SLAPI_LOG_FATAL, "krb5_modify","Could not find entry %s (error %d)\n",dn, rc );
	  slapi_send_ldap_result( pb, rc, NULL, NULL, 0, NULL );
	  return(1);
	}
	if ( e != NULL ) {
		Slapi_Value	*credval, **pwvals;
		int		i, hint, valcount;
                struct berval   **auth_idval;
		char principal[255];
                principal[0]=0;
                if ( slapi_entry_attr_find( e, "infnKerberosPrincipal", &attr ) == 0 ) {
                   slapi_attr_get_values( attr, &auth_idval );
                   strncpy(principal, auth_idval[0]->bv_val,254);
                }
                if (strlen(principal)) {
			LDAPMod         **mods;
        		char            *attr_name,*attr_value;
        		int             i,j;
        		char unhashedUserPassword[255],newPrincipal[255];
			int pwChanging;
        		unhashedUserPassword[0]=0;
        		newPrincipal[0]=0;
                        pwChanging=0;
        		slapi_pblock_get( pb, SLAPI_MODIFY_MODS, &mods );
  			for (i = 0; mods[i] ; i++) {
                		attr_name = mods[i]->mod_type;
                		char * config_attr_value;
                		for (j = 0; mods[i]->mod_bvalues && mods[i]->mod_bvalues[j]; j++) {
                       			attr_value = (char *) mods[i]->mod_bvalues[j]->bv_val;
                       			if (!unhashedUserPassword[0] && !strcmp(attr_name,"unhashed#user#password")) {
                          			strncpy(unhashedUserPassword,attr_value,254);
						pwChanging=1;
                       			}
                       			if (!newPrincipal[0] && !strcmp(attr_name,"infnKerberosPrincipal")) {
                          			strncpy(newPrincipal,attr_value,254);
                       			}
 		       			//slapi_log_error( SLAPI_LOG_FATAL, "krb5_modify","%s=%s\n",attr_name,attr_value);
				}
			}
			if (newPrincipal[0]) {
                           strncpy(principal,newPrincipal,254);
                        }
   			char *charToUP = 0L;
   			if((charToUP = strstr (principal, "@")) && charToUP ) {
     				charToUP++;
     				for(; *charToUP; charToUP++) {
         				*charToUP = toupper(*charToUP);
     				}
   			}
			if (FALSE) {
//			if (pwChanging) {
			  unsigned long krb5_perror;
                          slapi_log_error( SLAPI_LOG_FATAL, "krb5_modify", "Trying to change the password for the principal %s\n", principal);
			  krb5_perror = k5_cpw(principal, unhashedUserPassword);
			  if (krb5_perror){
		  	  	slapi_log_error( SLAPI_LOG_FATAL, "krb5_modify", "ERROR: Password has not been changed for the principal %s\n", principal);
			  } else {
		  	  	slapi_log_error( SLAPI_LOG_FATAL, "krb5_modify", "SUCCESS: Password has been changed for the principal %s\n", principal);
			  }
			  if (!(int)(krb5_perror-43787532UL)) {
                                slapi_log_error( SLAPI_LOG_FATAL, "krb5_modify", "Trying to add the principal %s\n", principal);
				krb5_perror = k5_addprinc(principal,unhashedUserPassword);
				if (krb5_perror){
			  	   slapi_log_error( SLAPI_LOG_FATAL, "krb5_modify", "ERROR: Principal %s has not been added.\n", principal);
				} else {
			  	    slapi_log_error( SLAPI_LOG_FATAL, "krb5_modify", "SUCCESS: Principal %s has been added.\n", principal);
				}
			  }
			  if (krb5_perror) {
                                slapi_send_ldap_result( pb, LDAP_UNWILLING_TO_PERFORM,NULL, NULL, 0, NULL );
				return(-1);
			  }
			}
                }
        }
	return(0);
}


int krb5plugin_init( Slapi_PBlock *pb )
{

        slapi_log_error( SLAPI_LOG_FATAL, "krb5plugin_init",
                        "Kerberos 5 bind pre-operation plugin (INFN) initializing\n" );
	/* Retrieve and save the plugin identity to later pass to
	   internal operations */
	if ( slapi_pblock_get( pb, SLAPI_PLUGIN_IDENTITY, &plugin_id ) != 0 ) {
		slapi_log_error( SLAPI_LOG_FATAL, "krb5plugin_init",
			"Failed to retrieve SLAPI_PLUGIN_IDENTITY\n" );
		return( -1 );
	}

	/* Register the pre-operation bind function and specify
	   the server plug-in version. */
	if ( slapi_pblock_set( pb, SLAPI_PLUGIN_VERSION, 
		SLAPI_PLUGIN_VERSION_01 ) != 0 ||
	     slapi_pblock_set( pb, SLAPI_PLUGIN_DESCRIPTION, 
		(void *)&bindpdesc ) != 0 ||
	     slapi_pblock_set( pb, SLAPI_PLUGIN_PRE_BIND_FN, 
		(void *) krb5_bind ) != 0 ) {
	     //slapi_pblock_set( pb, SLAPI_PLUGIN_PRE_MODIFY_FN, 
	     //(void *) krb5_modify ) != 0 ) {

		slapi_log_error( SLAPI_LOG_FATAL, "krb5plugin_init",
			"Failed to set version and functions\n" );
		return( -1 );
	}

	return( 0 );
}

